@@ -14,15 +14,25 @@
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
+
+#ifndef _LIBNFS_PRIVATE_H_
+#define _LIBNFS_PRIVATE_H_
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"  /* HAVE_SOCKADDR_STORAGE ? */
 #endif
 
+#ifndef WIN32
 #include <sys/socket.h>  /* struct sockaddr_storage */
+#endif
 
 #include "libnfs-zdr.h"
 
-#ifndef HAVE_SOCKADDR_STORAGE
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+#if !defined(HAVE_SOCKADDR_STORAGE) && !defined(WIN32)
 /*
  * RFC 2553: protocol-independent placeholder for socket addresses
  */
@@ -53,10 +63,23 @@ struct rpc_fragment {
 };
 
 #define RPC_CONTEXT_MAGIC 0xc6e46435
+#define RPC_PARAM_UNDEFINED -1
+
+/*
+ * Queue is singly-linked but we hold on to the tail
+ */
+struct rpc_queue {
+	struct rpc_pdu *head, *tail;
+};
+
+#define HASHES 1024
+#define NFS_RA_TIMEOUT 5
+#define NFS_MAX_XFER_SIZE (1024 * 1024)
 
 struct rpc_context {
-        uint32_t magic;
+	uint32_t magic;
 	int fd;
+	int old_fd;
 	int is_connected;
 
 	char *error_string;
@@ -65,37 +88,44 @@ struct rpc_context {
 	void *connect_data;
 
 	struct AUTH *auth;
-	unsigned long xid;
+	uint32_t xid;
 
-       /* buffer used for encoding RPC PDU */
-       char *encodebuf;
-       int encodebuflen;
+	/* buffer used for encoding RPC PDU */
+	char *encodebuf;
+	int encodebuflen;
 
-       struct rpc_pdu *outqueue;
-       struct sockaddr_storage udp_src;
-       struct rpc_pdu *waitpdu;
+	struct rpc_queue outqueue;
+	struct sockaddr_storage udp_src;
+	struct rpc_queue waitpdu[HASHES];
 
-       uint32_t inpos;
-       uint32_t insize;
-       char *inbuf;
+	uint32_t inpos;
+	uint32_t insize;
+	char *inbuf;
 
-       /* special fields for UDP, which can sometimes be BROADCASTed */
-       int is_udp;
-       struct sockaddr *udp_dest;
-       int is_broadcast;
+	/* special fields for UDP, which can sometimes be BROADCASTed */
+	int is_udp;
+	struct sockaddr *udp_dest;
+	int is_broadcast;
 
-       /* track the address we connect to so we can auto-reconnect on session failure */
-       struct sockaddr_storage s;
-       int auto_reconnect;
+	/* track the address we connect to so we can auto-reconnect on session failure */
+	struct sockaddr_storage s;
+	int auto_reconnect;
 
 	/* fragment reassembly */
 	struct rpc_fragment *fragments;
+
+	/* parameters passable via URL */
+	int tcp_syncnt;
+	int uid;
+	int gid;
+	uint32_t readahead;
+	int debug;
 };
 
 struct rpc_pdu {
 	struct rpc_pdu *next;
 
-	unsigned long xid;
+	uint32_t xid;
 	ZDR zdr;
 
 	uint32_t written;
@@ -110,15 +140,36 @@ struct rpc_pdu {
 	uint32_t zdr_decode_bufsize;
 };
 
+void rpc_reset_queue(struct rpc_queue *q);
+void rpc_enqueue(struct rpc_queue *q, struct rpc_pdu *pdu);
+void rpc_return_to_queue(struct rpc_queue *q, struct rpc_pdu *pdu);
+unsigned int rpc_hash_xid(uint32_t xid);
+
 struct rpc_pdu *rpc_allocate_pdu(struct rpc_context *rpc, int program, int version, int procedure, rpc_cb cb, void *private_data, zdrproc_t zdr_decode_fn, int zdr_bufsize);
 void rpc_free_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu);
 int rpc_queue_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu);
 int rpc_get_pdu_size(char *buf);
 int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size);
-void rpc_error_all_pdus(struct rpc_context *rpc, char *error);
+void rpc_error_all_pdus(struct rpc_context *rpc, const char *error);
 
-void rpc_set_error(struct rpc_context *rpc, char *error_string, ...);
-void nfs_set_error(struct nfs_context *nfs, char *error_string, ...);
+void rpc_set_error(struct rpc_context *rpc, const char *error_string, ...)
+#ifdef __GNUC__
+ __attribute__((format(printf, 2, 3)))
+#endif
+;
+
+void nfs_set_error(struct nfs_context *nfs, char *error_string, ...)
+#ifdef __GNUC__
+ __attribute__((format(printf, 2, 3)))
+#endif
+;
+
+#define RPC_LOG(rpc, level, format, ...) \
+	do { \
+		if (level <= rpc->debug) { \
+			fprintf(stderr, "libnfs:%d " format "\n", level, ## __VA_ARGS__); \
+		} \
+	} while (0)
 
 const char *nfs_get_server(struct nfs_context *nfs);
 const char *nfs_get_export(struct nfs_context *nfs);
@@ -132,7 +183,19 @@ struct sockaddr *rpc_get_recv_sockaddr(struct rpc_context *rpc);
 void rpc_set_autoreconnect(struct rpc_context *rpc);
 void rpc_unset_autoreconnect(struct rpc_context *rpc);
 
+void rpc_set_tcp_syncnt(struct rpc_context *rpc, int v);
+void rpc_set_uid(struct rpc_context *rpc, int uid);
+void rpc_set_gid(struct rpc_context *rpc, int gid);
+void rpc_set_readahead(struct rpc_context *rpc, uint32_t v);
+void rpc_set_debug(struct rpc_context *rpc, int level);
+
 int rpc_add_fragment(struct rpc_context *rpc, char *data, uint64_t size);
 void rpc_free_all_fragments(struct rpc_context *rpc);
 
 const struct nfs_fh3 *nfs_get_rootfh(struct nfs_context *nfs);
+
+#ifdef __cplusplus
+}
+#endif
+
+#endif /* !_LIBNFS_PRIVATE_H_ */
