@@ -132,6 +132,10 @@ int rpc_get_fd(struct rpc_context *rpc)
 {
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
+	if (rpc->old_fd) {
+		return rpc->old_fd;
+	}
+
 	return rpc->fd;
 }
 
@@ -200,6 +204,7 @@ static int rpc_write_to_socket(struct rpc_context *rpc)
 	return 0;
 }
 
+#define MAX_UDP_SIZE 65536
 static int rpc_read_from_socket(struct rpc_context *rpc)
 {
 	int size;
@@ -209,105 +214,158 @@ static int rpc_read_from_socket(struct rpc_context *rpc)
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
 	if (rpc->is_udp) {
-		int available;
+		char *buf;
 		socklen_t socklen = sizeof(rpc->udp_src);
 
-		if (ioctl(rpc->fd, FIONREAD, &available) != 0) {
-			rpc_set_error(rpc, "Ioctl FIONREAD returned error : %d. Closing socket.", errno);
+		buf = malloc(MAX_UDP_SIZE);
+		if (buf == NULL) {
+			rpc_set_error(rpc, "Failed to malloc buffer for recvfrom");
 			return -1;
 		}
-
-		if (available == 0) {
-			rpc_set_error(rpc, "Socket has been closed");
-			return -1;
-		}
-
-		if (available > rpc->inbuflen) {
-			rpc_set_error(rpc, "RPC fragment too big");
-			return -1;
-		}
-
-		count = recvfrom(rpc->fd, rpc->inbuf, available, MSG_DONTWAIT, (struct sockaddr *)&rpc->udp_src, &socklen);
-		if (count < 0) {
+		count = recvfrom(rpc->fd, buf, MAX_UDP_SIZE, MSG_DONTWAIT, (struct sockaddr *)&rpc->udp_src, &socklen);
+		if (count == -1) {
+			free(buf);
+			if (errno == EINTR || errno == EAGAIN) {
+				return 0;
+			}
 			rpc_set_error(rpc, "Failed recvfrom: %s", strerror(errno));
 			return -1;
 		}
-		if (rpc_process_pdu(rpc, rpc->inbuf, count) != 0) {
+		if (rpc_process_pdu(rpc, buf, count) != 0) {
 			rpc_set_error(rpc, "Invalid/garbage pdu received from server. Ignoring PDU");
+			free(buf);
 			return -1;
 		}
+		free(buf);
 		return 0;
 	}
 
-	size = rpc->inbuflen - rpc->inpos;
+	/* read record marker, 4 bytes at the beginning of every pdu */
+	if (rpc->inbuf == NULL) {
+		rpc->insize = 4;
+		rpc->inbuf = malloc(rpc->insize);
+		if (rpc->inbuf == NULL) {
+			rpc_set_error(rpc, "Failed to allocate buffer for record marker, errno:%d. Closing socket.", errno);
+			return -1;
+		}
+	}
 
+	if (rpc->inpos < 4) {
+		size = 4 - rpc->inpos;
+
+		count = recv(rpc->fd, rpc->inbuf + rpc->inpos, size, MSG_DONTWAIT);
+		if (count == -1) {
+			if (errno == EINTR || errno == EAGAIN) {
+				return 0;
+			}
+			rpc_set_error(rpc, "Read from socket failed, errno:%d. Closing socket.", errno);
+			return -1;
+		}
+		rpc->inpos += count;
+
+		if (rpc->inpos < 4) {
+			return 0;
+		}
+	}
+
+	pdu_size = rpc_get_pdu_size(rpc->inbuf);
+
+	if (pdu_size > NFS_MAX_XFER_SIZE + 4096) {
+		rpc_set_error(rpc, "Incoming PDU exceeds limit of %d bytes.", NFS_MAX_XFER_SIZE + 4096);
+		return -1;
+	}
+
+	if (rpc->insize < pdu_size) {
+		rpc->inbuf = realloc(rpc->inbuf, pdu_size);
+		if (rpc->inbuf == NULL) {
+			rpc_set_error(rpc, "Failed to allocate buffer of %d bytes for pdu, errno:%d. Closing socket.", pdu_size, errno);
+			return -1;
+		}
+		rpc->insize = pdu_size;
+	}
+
+	size = rpc->insize - rpc->inpos;
 	count = recv(rpc->fd, rpc->inbuf + rpc->inpos, size, MSG_DONTWAIT);
-	if (count < 0) {
+	if (count == -1) {
 		if (errno == EINTR || errno == EAGAIN) {
 			return 0;
 		}
 		rpc_set_error(rpc, "Read from socket failed, errno:%d. Closing socket.", errno);
 		return -1;
 	}
+
+	if (count == 0) {
+		/* remote side has closed the socket. Reconnect. */
+		return -1;
+	}
+
 	rpc->inpos += count;
 
-	while (rpc->inpos >= 4) {
-		pdu_size = rpc_get_pdu_size(rpc->inbuf);
-		if (pdu_size > NFS_MAX_XFER_SIZE + 4096) {
-			rpc_set_error(rpc, "Incoming PDU exceeds limit of %d bytes.", NFS_MAX_XFER_SIZE + 4096);
+	if (rpc->inpos == rpc->insize) {
+		char *buf = rpc->inbuf;
+
+		rpc->inbuf  = NULL;
+		rpc->insize = 0;
+		rpc->inpos  = 0;
+
+		if (rpc_process_pdu(rpc, buf, pdu_size) != 0) {
+			rpc_set_error(rpc, "Invalid/garbage pdu received from server. Closing socket");
 			return -1;
 		}
-
-		if (rpc->inpos >= pdu_size) {
-			if (rpc_process_pdu(rpc, rpc->inbuf, pdu_size) != 0) {
-				rpc_set_error(rpc, "Invalid/garbage pdu received from server. Closing socket");
-				return -1;
-			}
-			memmove(rpc->inbuf, rpc->inbuf + pdu_size, rpc->inpos - pdu_size);
-			rpc->inpos -= pdu_size;
-			continue;
-		}
-		break;
-	} 
+		free(buf);
+	}
 
 	return 0;
 }
 
 
+static void
+maybe_call_connect_cb(struct rpc_context *rpc, int status)
+{
+	rpc_cb tmp_cb = rpc->connect_cb;
+
+	if (rpc->connect_cb == NULL) {
+		return;
+	}
+
+	rpc->connect_cb = NULL;
+	tmp_cb(rpc, status, rpc->error_string, rpc->connect_data);
+}
+
 int rpc_service(struct rpc_context *rpc, int revents)
 {
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
-	if (revents & POLLERR) {
+	if (revents & (POLLERR|POLLHUP)) {
+		if (revents & POLLERR) {
 #ifdef WIN32
-		char err = 0;
+			char err = 0;
 #else
-		int err = 0;
+			int err = 0;
 #endif
-		socklen_t err_size = sizeof(err);
+			socklen_t err_size = sizeof(err);
 
-		if (getsockopt(rpc->fd, SOL_SOCKET, SO_ERROR,
+			if (getsockopt(rpc->fd, SOL_SOCKET, SO_ERROR,
 				(char *)&err, &err_size) != 0 || err != 0) {
-			if (err == 0) {
-				err = errno;
+				if (err == 0) {
+					err = errno;
+				}
+				rpc_set_error(rpc, "rpc_service: socket error "
+						    "%s(%d).",
+						    strerror(err), err);
+			} else {
+				rpc_set_error(rpc, "rpc_service: POLLERR, "
+						   "Unknown socket error.");
 			}
-			rpc_set_error(rpc, "rpc_service: socket error "
-					       "%s(%d).",
-					       strerror(err), err);
-		} else {
-			rpc_set_error(rpc, "rpc_service: POLLERR, "
-						"Unknown socket error.");
 		}
-		if (rpc->connect_cb != NULL) {
-			rpc->connect_cb(rpc, RPC_STATUS_ERROR, rpc->error_string, rpc->connect_data);
-		}
-		return -1;
-	}
-	if (revents & POLLHUP) {
-		rpc_set_error(rpc, "Socket failed with POLLHUP");
-		if (rpc->connect_cb != NULL) {
-			rpc->connect_cb(rpc, RPC_STATUS_ERROR, rpc->error_string, rpc->connect_data);
+		if (revents & POLLHUP) {
+			rpc_set_error(rpc, "Socket failed with POLLHUP");
 		}
+		if (rpc->auto_reconnect) {
+			return rpc_reconnect_requeue(rpc);
+		}
+		maybe_call_connect_cb(rpc, RPC_STATUS_ERROR);
 		return -1;
+
 	}
 
 	if (rpc->is_connected == 0 && rpc->fd != -1 && revents&POLLOUT) {
@@ -322,32 +380,25 @@ int rpc_service(struct rpc_context *rpc, int revents)
 			rpc_set_error(rpc, "rpc_service: socket error "
 				  	"%s(%d) while connecting.",
 					strerror(err), err);
-			if (rpc->connect_cb != NULL) {
-				rpc->connect_cb(rpc, RPC_STATUS_ERROR,
-					NULL, rpc->connect_data);
-			}
+			maybe_call_connect_cb(rpc, RPC_STATUS_ERROR);
 			return -1;
 		}
 
 		rpc->is_connected = 1;
-		RPC_LOG(rpc, 2, "connection established");
-		if (rpc->connect_cb != NULL) {
-			rpc->connect_cb(rpc, RPC_STATUS_SUCCESS, NULL, rpc->connect_data);
-		}
+		RPC_LOG(rpc, 2, "connection established on fd %d", rpc->fd);
+		maybe_call_connect_cb(rpc, RPC_STATUS_SUCCESS);
 		return 0;
 	}
 
 	if (revents & POLLIN) {
 		if (rpc_read_from_socket(rpc) != 0) {
-		  	rpc_reconnect_requeue(rpc);
-			return 0;
+			return rpc_reconnect_requeue(rpc);
 		}
 	}
 
 	if (revents & POLLOUT && rpc_has_queue(&rpc->outqueue)) {
 		if (rpc_write_to_socket(rpc) != 0) {
-			rpc_set_error(rpc, "write to socket failed");
-			return -1;
+			return rpc_reconnect_requeue(rpc);
 		}
 	}
 
@@ -414,6 +465,14 @@ static int rpc_connect_sockaddr_async(struct rpc_context *rpc, struct sockaddr_s
 		return -1;
 	}
 
+	if (rpc->old_fd) {
+		if (dup2(rpc->fd, rpc->old_fd) == -1) {
+			return -1;
+		}
+		close(rpc->fd);
+		rpc->fd = rpc->old_fd;
+	}
+
 	/* Some systems allow you to set capabilities on an executable
 	 * to allow the file to be executed with privilege to bind to
 	 * privileged system ports, even if the user is not root.
@@ -568,28 +627,29 @@ static void reconnect_cb(struct rpc_context *rpc, int status, void *data _U_, vo
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
 	if (status != RPC_STATUS_SUCCESS) {
-		rpc_error_all_pdus(rpc, "RPC ERROR: Failed to reconnect async");
+		rpc_set_error(rpc, "Failed to reconnect async");
+		rpc_reconnect_requeue(rpc);
 		return;
 	}
 
 	rpc->is_connected = 1;
 	rpc->connect_cb   = NULL;
+	rpc->old_fd = 0;
 }
 
 /* disconnect but do not error all PDUs, just move pdus in-flight back to the outqueue and reconnect */
 static int rpc_reconnect_requeue(struct rpc_context *rpc)
 {
-	struct rpc_pdu *pdu;
+	struct rpc_pdu *pdu, *next;
 	unsigned int i;
 
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
 	if (rpc->fd != -1) {
-		close(rpc->fd);
+		rpc->old_fd = rpc->fd;
 	}
 	rpc->fd  = -1;
 	rpc->is_connected = 0;
-	rpc->inpos = 0;
 
 	if (rpc->outqueue.head) {
 		rpc->outqueue.head->written = 0;
@@ -600,8 +660,8 @@ static int rpc_reconnect_requeue(struct rpc_context *rpc)
 	 */
 	for (i = 0; i < HASHES; i++) {
 		struct rpc_queue *q = &rpc->waitpdu[i];
-
-		for (pdu=q->head; pdu; pdu=pdu->next) {
+		for (pdu = q->head; pdu; pdu = next) {
+			next = pdu->next;
 			rpc_return_to_queue(&rpc->outqueue, pdu);
 			/* we have to re-send the whole pdu again */
 			pdu->written = 0;
@@ -618,6 +678,7 @@ static int rpc_reconnect_requeue(struct rpc_context *rpc)
 		}
 	} else {
 		RPC_LOG(rpc, 1, "reconnect NOT initiated, auto-reconnect is disabled");
+		return -1;
 	}
 
 	return 0;
